package flink.zanxiangnet.ad.monitoring.sink;

import com.aliyun.odps.tunnel.TableTunnel;
import com.aliyun.odps.tunnel.TunnelException;
import flink.zanxiangnet.ad.monitoring.config.ClickhouseDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.maxcompute.bean.BeanUtil;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.sink.RichSinkFunction;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

/**
 * 批量数据写出
 *
 * @param <IN>
 */
@Slf4j
public class ClickhouseBatchStreamSink<T, IN extends List<T>> extends RichSinkFunction<IN> {

    // 对象锁，防止MaxCompute的 Tunnel对象多次初始化
    private static final Object DUMMY_LOCK = new Object();

    private String projectName;
    private String tableName;

    private volatile transient TableTunnel tunnel;
    private volatile transient List<BeanUtil.FieldInfo> fieldInfoList;
    private volatile transient Map<String, Method> partitionFieldMethods;
    private SqlSessionFactory sqlSessionFactory;


    @Override
    public void open(Configuration config) throws Exception {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties ckProps = new Properties();
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_URL, params.get(ApplicationProperties.CK_URL));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_USER, params.get(ApplicationProperties.CK_USERNAME));
        ckProps.setProperty(ClickhouseDataSourceFactory.PROP_PASSWORD, params.get(ApplicationProperties.CK_PASSWORD));

        DataSourceFactory dataSourceFactory = new ClickhouseDataSourceFactory();
        dataSourceFactory.setProperties(ckProps);
        Environment environment = new Environment("clickhouse", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);
    }

    /**
     * 将值写入到 Sink。每个值都会调用此函数
     *
     * @param value
     * @param context
     */
    @Override
    public void invoke(IN value, Context context) throws TunnelException, IOException, InvocationTargetException, IllegalAccessException {
        AdStatOfDayDWD adStatOfDayDWD = new AdStatOfDayDWD();
        adStatOfDayDWD.setStatDay("2021-12-01");
        adStatOfDayDWD.setAccountId(100L);
        adStatOfDayDWD.setCreateTime(new Date());
        adStatOfDayDWD.setCostDeviationRateTotal(39.5);
        adStatOfDayDWD.removeNull();

        List<AdStatOfDayDWD> list = new ArrayList<>(6);
        for (int i = 0; i < 3; i ++) {
            AdStatOfDayDWD temp = new AdStatOfDayDWD();
            BeanUtils.copyProperties(adStatOfDayDWD, temp);
            temp.setAdId((long) i);
            list.add(temp);
        }
        list.add(adStatOfDayDWD);
        log.error("准备写入：{}", value.size());
        try (SqlSession session = sqlSessionFactory.openSession()) {
            AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
            mapper.add(adStatOfDayDWD);
            mapper.addBatch(list);
            List<AdStatOfDayDWD> all = mapper.selectAll();
            log.error("查询结果：{} | {}", all.size(), JsonUtil.toString(list));
        }
    }

    @Override
    public void close() throws Exception {
        super.close();
    }

    private String generateSQL(T t) {
        if (CollectionUtils.isEmpty(partitionFieldMethods)) {
            return null;
        }
        StringBuilder partition = new StringBuilder();
        for (Map.Entry<String, Method> entry : partitionFieldMethods.entrySet()) {
            partition.append(entry.getKey()).append("=");
            try {
                partition.append(entry.getValue().invoke(t));
            } catch (InvocationTargetException | IllegalAccessException e) {
                // 获取分区字段的值失败
                log.error(e.getMessage(), e);
                throw new RuntimeException("Failed get partition field value!");
            }
            partition.append(",");
        }
        partition = new StringBuilder(partition.substring(0, partition.length() - 1));
        return partition.toString();
    }
}
